program mag_distortion_correct

    use Globals


    implicit none

    call main()

    contains

    ! The actual subroutine
    subroutine main()

        use UserSuppliedParameters
        use UserInputs
        use ImageFiles
        use Images
        use ProgressBars
        use UsefulFunctions
        use Units
        use StringManipulations, only : IntegerToString

        ! Variables associated with user input

        type(UserInput)            ::   my_user_input
        type(UserSuppliedFilename) ::   input_filename
        type(UserSuppliedFilename) ::   output_filename
        type(UserSuppliedReal)     ::   distortion_angle
        type(UserSuppliedReal)     ::   x_scale
        type(UserSuppliedReal)     ::   y_scale
        type(UserSuppliedLogical)  ::   do_gain_correction
        type(UserSuppliedFilename) ::   gain_reference_filename
        type(UserSuppliedLogical)  ::   do_resampling
        type(UserSuppliedInteger)  ::   new_x_dimension
        type(UserSuppliedInteger)  ::   new_y_dimension

        ! Other
        type(ImageFile)            ::  input_imagefile
        type(Image)                ::  input_image
        type(Image)                ::  output_image
        type(Image)                ::  gain_reference
        type(ProgressBar)          ::  my_progress_bar

        integer                    ::  number_of_input_images
        integer                    ::  image_counter

        integer                    ::  i
        integer                    ::  j

        integer                    ::  stack_dimensions(3)





        ! Initialise this_program with the program name and version number
        call this_program%Init('mag_distortion_correct', '1.0.0','2015')


        ! Do the user input
        call my_user_input%Init(this_program%program_name)

        input_filename                  =   my_user_input%GetFilenameFromUser('Input stack filename', &
                                       'The input file, containing your original uncorrected images', &
                                       'input_filename', 'my_image_stack.mrc', .true.)

        ! Open the input file
        call input_imagefile%Init(input_filename%value)

        output_filename                 =   my_user_input%GetFilenameFromUser('Output stack filename', &
                                       'The output file, containing the corrected_images', &
                                       'output_filename', 'my_image_stack_corrected.mrc', .false.)


        distortion_angle                =   my_user_input%GetRealFromUser('Distortion Angle (Degrees)', &
                                       'Angle of the distortion', 'distortion_angle', '0')

        x_scale                         =   my_user_input%GetRealFromUser('Major scale', &
                                       'Scale factor along major axis', 'major_scale', '1', min_value = 0.0e0)

        y_scale                         =   my_user_input%GetRealFromUser('Minor scale', &
                                       'Scale factor along minor axis', 'minor_scale', '1', min_value = 0.0e0)

        do_gain_correction              =   my_user_input%GetLogicalFromUser('Do gain correction before undistorting?', &
                                        'If yes, a gain reference will be applied to the input images', &
                                        'do_gain_correction', 'no')

        if (do_gain_correction%value) then

            gain_reference_filename     =   my_user_input%GetFilenameFromUser('Gain reference filename', &
                                        'Filename of gain reference, should be MRC', &
                                        'gain_reference_filename', 'gain_reference.mrc', .true.)
        endif

        do_resampling                   =   my_user_input%GetLogicalFromUser('Resample images?', &
                                        'Resample images after distortion correction and gain correction, by cropping '//&
                                        'their Fourier transforms', &
                                        'do_resampling','no' )

        if (do_resampling%value) then
            new_x_dimension             =   my_user_input%GetIntegerFromUser(   'New X dimension', &
                                        'X dimension of output image(s)', 'new_x_dimension', &
                                        IntegerToString(int(input_imagefile%GetDimension(1)/2)))
            new_y_dimension             =   my_user_input%GetIntegerFromUser(   'New Y dimension', &
                                        'Y dimension of output image(s)', 'new_y_dimension', &
                                        IntegerToString(int(input_imagefile%GetDimension(2)/2)))
        endif

        call my_user_input%UpdateDefaults()
        write(*,*)

        ! Get dimensions of input
        number_of_input_images = input_imagefile%GetStackSize()
        stack_dimensions = input_imagefile%GetDimensions();



        ! Read in the gain reference if needed
        if (do_gain_correction%value) then
            call gain_reference%ReadFromDisk(gain_reference_filename%value)
        endif


        !
        ! Loop over images
        !
        call my_progress_bar%Begin(number_of_input_images)
        do image_counter = 1, number_of_input_images

            ! Read image from disk
            call input_image%ReadFromDisk(input_imagefile, image_counter)

            ! Reset pixel values of output image
            call output_image%Allocate(dims=[stack_dimensions(1), stack_dimensions(2), 1],in_real_space=.true.)
            output_image = 0.0e0

            ! Apply gain to input image
            if (do_gain_correction%value) then
                !$omp parallel default(shared) private(i,j)
                !$omp do
                do j=1,input_image%GetLogicalDimension(2)
                    do i=1,input_image%GetLogicalDimension(1)
                        input_image%real_values(i,j,1) = 16.0 * gain_reference%real_values(i,j,1) * input_image%real_values(i,j,1)
                    enddo
                enddo
                !$omp enddo
                !$omp end parallel
            endif

            ! Do correction for magnification distortion
            call input_image%CorrectMagnificationDistortion(output_image,distortion_angle%value,x_scale%value,y_scale%value)

            ! Do resampling
            if (do_resampling%value) then
                call output_image%ForwardFFT()
                call input_image%Allocate(dims=[new_x_dimension%value,new_y_dimension%value,1],in_real_space=.false.)
                call output_image%ClipInto(input_image)
                call input_image%BackwardFFT()
                output_image = input_image
            endif

            ! Write out the result
            call output_image%WriteToDisk(output_filename%value, image_counter)
            call my_progress_bar%Update(image_counter)

        enddo ! End of loop over images

        call my_progress_bar%Finish()

        ! Deallocation
        call input_imagefile%Close()

        ! All done
        call this_program%Terminate()

    end subroutine main

end program mag_distortion_correct
